/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.fleet.implement;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import nl.cloudfarming.client.farm.model.FarmManager;
import nl.cloudfarming.client.fleet.model.Implement;
import nl.cloudfarming.client.fleet.model.ImplementManager;
import nl.cloudfarming.client.fleet.model.Machine;
import nl.cloudfarming.client.fleet.model.MachineManager;
import nl.cloudfarming.client.fleet.model.MachineProvider;
import nl.cloudfarming.client.model.AbstractEntityManager;
import nl.cloudfarming.client.model.AgroURI;
import nl.cloudfarming.client.model.EntityManager;
import nl.cloudfarming.client.model.ItemIdType;
import org.netbeans.api.project.Project;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.util.Exceptions;
import org.openide.util.Lookup;
import org.openide.util.lookup.ServiceProvider;
import org.openide.util.lookup.ServiceProviders;

/**
 *
 * @author Wytse Visser
 */
@ServiceProviders({
    @ServiceProvider(service = ImplementManager.class),
    @ServiceProvider(service = EntityManager.class)
})
public class ImplementManagerImpl extends AbstractEntityManager<Implement> implements ImplementManager, MachineManager<Implement>, MachineProvider {

    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.fleet.implement");
    private static final String FLEET_DIR = "Fleet";
    private static final String IMPLEMENTS_DIR = "Implements";

    @Override
    public Implement load(FileObject fileObject) {
        Implement implement = null;
        try (ObjectInputStream ois = new ObjectInputStream(fileObject.getInputStream())) {
            implement = (Implement) ois.readObject();
        } catch (ClassNotFoundException | IOException | NullPointerException ex) {
            LOGGER.log(Level.WARNING, "Could not create implement from fileObject", ex.getMessage());
        }
        return getCachedInstanceFor(implement);
    }

    @Override
    public void save(Implement implement) {
        try {
            FarmManager farmManager = Lookup.getDefault().lookup(FarmManager.class);
            Project project = farmManager.findFarmProject(implement.getURI());
            FileObject implementFolder = getImplementsSubfolder(project, true);
            FileObject implementObject = implementFolder.getFileObject(implement.getURI().getItemIdNumber(), Implement.EXT);
            
            if (implementObject == null) {
                implementObject = implementFolder.createData(implement.getURI().getItemIdNumber(), Implement.EXT);
            }
            save(implement, implementObject);

        } catch (IOException ex) {
            LOGGER.log(Level.SEVERE, "Error while trying to create file object for Implement {0}", ex.getMessage());
            Exceptions.printStackTrace(ex);

        }
    }

    private void save(Implement implement, FileObject fileObject) {
        try (ObjectOutputStream oos = new ObjectOutputStream(fileObject.getOutputStream())) {
            oos.writeObject(implement);
            addToCache(implement);
        } catch (IOException ex) {
            Exceptions.printStackTrace(ex);
        }
    }

    /**
     * Method that finds all Implements of a farm. It loops through all the
     * FileObjects contained by the implements folder, gets a DataObject for
     * each FileObject, fetches the Implement object from the DataObject and
     * uses the find method of the AbstractEntityManager to either fetch the
     * Implement from the cache or load it using the proper channels. Using the
     * cache prevents Implements from being loaded multiple times.
     *
     * @param project The project you wish to fetch all Implements from
     * @return List containing all the Implements of the project passed as
     * parameter
     */
    @Override
    public List<Machine> findAll(Project project) {
        List<Machine> impls = new ArrayList<>();
        FileObject implementsDir = getImplementsSubfolder(project, true);

        //for each serialized vehicle file
        for (FileObject implementFile : implementsDir.getChildren()) {
            try {
                //get the DataObject for the file
                DataObject dataObject = DataObject.find(implementFile);
                if (dataObject != null) {
                    //get the Implement contained by the DataObject
                    Implement implement = dataObject.getLookup().lookup(Implement.class);
                    if (implement != null) {
                        //use the AbstractEntityManager to get the Implement from the cache
                        //if possible
                        impls.add(find(implement.getURI()));
                    }
                }
            } catch (DataObjectNotFoundException ex) {
                LOGGER.log(Level.WARNING, "Could not find DataObject for file {0} {1}", new Object[]{implementFile, ex.getMessage()});
            }
        }

        return impls;
    }

    private static FileObject getFleetFolder(Project project, boolean create) {
        FileObject farmFolder = project.getProjectDirectory();
        FileObject fleetFolder = farmFolder.getFileObject(FLEET_DIR);

        if (fleetFolder == null && create) {
            try {
                fleetFolder = farmFolder.createFolder(FLEET_DIR);
            } catch (IOException ex) {
                LOGGER.log(Level.SEVERE, "Could not create the folder {0} in folder {1}", new Object[]{FLEET_DIR, farmFolder.getPath()});
            }
        }

        return fleetFolder;
    }

    private static FileObject getImplementsSubfolder(Project project, boolean create) {
        FileObject fleetFolder = getFleetFolder(project, create);
        FileObject implementsFolder = fleetFolder.getFileObject(IMPLEMENTS_DIR);

        // Create folder if it doesn't exist
        if (implementsFolder == null && create) {
            try {
                implementsFolder = fleetFolder.createFolder(IMPLEMENTS_DIR);
            } catch (IOException ex) {
                LOGGER.log(Level.SEVERE, "Could not create the folder {0} in folder {1}", new Object[]{IMPLEMENTS_DIR, fleetFolder.getPath()});
            }
        }
        return implementsFolder;
    }

    @Override
    public Set<ItemIdType> getSupportedTypes() {
        return EnumSet.of(ItemIdType.IMT);
    }

    @Override
    protected Implement load(AgroURI uri, Project project) {
        switch (uri.getItemIdType()) {
            case IMT:
                FileObject implementFile = getImplementsSubfolder(project, true).getFileObject(uri.getItemIdNumber(), Implement.EXT);
                return load(implementFile);
            default:
                LOGGER.log(Level.WARNING, "cannot load unsupported entity type {0}", new Object[]{uri.getItemIdType()});
                return null;
        }
    }

    @Override
    public void remove(Implement entity) {
        //FIXME implement remove
        throw new UnsupportedOperationException("Not supported yet.");
    }
}
